package com.acme.boot.autoconfigure.orm.hibernatereactive;

import io.smallrye.mutiny.Uni;
import org.hibernate.reactive.mutiny.Mutiny;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.AutoConfigurationPackages;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.boot.autoconfigure.domain.EntityScanPackages;
import org.springframework.boot.context.properties.ConfigurationProperties;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PathMatchingResourcePatternResolver;
import org.springframework.core.io.support.ResourcePatternResolver;
import org.springframework.core.type.classreading.CachingMetadataReaderFactory;
import org.springframework.core.type.classreading.MetadataReader;
import org.springframework.core.type.classreading.MetadataReaderFactory;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.persistence.*;
import javax.persistence.spi.ClassTransformer;
import javax.persistence.spi.PersistenceProvider;
import javax.persistence.spi.PersistenceUnitInfo;
import javax.persistence.spi.PersistenceUnitTransactionType;
import javax.sql.DataSource;
import java.net.URL;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Configuration
public class HibernateReactiveAutoConfiguration {

    @ConfigurationProperties(prefix = "spring.hibernate-reactive.properties")
    @Component
    static class HibernateReactiveProperties extends Properties {

        public HibernateReactiveProperties(Map<String, String> props) {
            putAll(props);
        }
    }

    @Bean(destroyMethod = "close")
    Mutiny.SessionFactory sessionFactory(BeanFactory beanFactory, HibernateReactiveProperties props) throws Exception {

        PersistenceProvider provider = new org.hibernate.reactive.provider.ReactivePersistenceProvider();

        var packages = EntityScanPackages.get(beanFactory).getPackageNames();
        if (packages.isEmpty() && AutoConfigurationPackages.has(beanFactory)) {
            packages = AutoConfigurationPackages.get(beanFactory);
        }

        List<String> classNames = packages.stream()
                .flatMap(p -> getMappedClasses(p).stream())
                .collect(Collectors.toList());

        PersistenceUnitInfo config = getPersistenceUnitInfo(provider, classNames, props);

        return provider.createContainerEntityManagerFactory(config, null).unwrap(Mutiny.SessionFactory.class);
    }

    private final PathMatchingResourcePatternResolver pathMatchingResourcePatternResolver = new PathMatchingResourcePatternResolver();
    private final MetadataReaderFactory metadataReaderFactory = new CachingMetadataReaderFactory();

    private List<String> getMappedClasses(String packageName) {
        String basePath = ResourcePatternResolver.CLASSPATH_ALL_URL_PREFIX
                + packageName.replaceAll("[.]", "/")
                + "/**/*.class";

        try {
            Resource[] resources = pathMatchingResourcePatternResolver.getResources(basePath);
            var classes = new ArrayList<String>();
            for (Resource resource : resources) {
                if (resource.isReadable()) {
                    MetadataReader metadataReader = metadataReaderFactory.getMetadataReader(resource);
                    Set<String> annotationTypes = metadataReader.getAnnotationMetadata().getAnnotationTypes();
                    if (Stream.of(Entity.class, MappedSuperclass.class).anyMatch(clazz -> annotationTypes.contains(clazz.getName()))) {
                        classes.add(metadataReader.getClassMetadata().getClassName());
                    }
                }
            }
            return classes;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private PersistenceUnitInfo getPersistenceUnitInfo(PersistenceProvider persistenceProvider, List<String> classNames, Properties properties) {
        return new PersistenceUnitInfo() {

            @Override
            public String getPersistenceUnitName() {
                return "";
            }

            @Override
            public String getPersistenceProviderClassName() {
                return persistenceProvider.getClass().getName();
            }

            @Override
            public PersistenceUnitTransactionType getTransactionType() {
                return PersistenceUnitTransactionType.RESOURCE_LOCAL;
            }

            @Override
            public DataSource getJtaDataSource() {
                return null;
            }

            @Override
            public DataSource getNonJtaDataSource() {
                return null;
            }

            @Override
            public List<String> getMappingFileNames() {
                return null;
            }

            @Override
            public List<URL> getJarFileUrls() {
                return null;
            }

            @Override
            public URL getPersistenceUnitRootUrl() {
                return null;
            }

            @Override
            public List<String> getManagedClassNames() {
                return classNames;
            }

            @Override
            public boolean excludeUnlistedClasses() {
                return false;
            }

            @Override
            public SharedCacheMode getSharedCacheMode() {
                return null;
            }

            @Override
            public ValidationMode getValidationMode() {
                return ValidationMode.AUTO;
            }

            @Override
            public Properties getProperties() {
                return properties;
            }

            @Override
            public String getPersistenceXMLSchemaVersion() {
                return null;
            }

            @Override
            public ClassLoader getClassLoader() {
                return getClass().getClassLoader();
            }

            @Override
            public void addTransformer(ClassTransformer transformer) {
            }

            @Override
            public ClassLoader getNewTempClassLoader() {
                return null;
            }
        };
    }

    @Component
    @ConditionalOnProperty(name = "spring.hibernate-reactive.load-second-level-cache-on-startup", matchIfMissing = true)
    static class SecondLevelCacheLoader {
        @Autowired
        private Mutiny.SessionFactory sessionFactory;

        private Logger logger = LoggerFactory.getLogger(getClass());

        @PostConstruct
        void loadCache() {
            for (var entityType : sessionFactory.getMetamodel().getEntities()) {
                Cacheable cacheable = entityType.getJavaType().getAnnotation(Cacheable.class);
                if (cacheable != null && cacheable.value()) {
                    sessionFactory.withSession(session -> {
                        String qlString = "select e from " + entityType.getName() + " e";
                        Uni<List<Object>> items = session.createQuery(qlString).getResultList();
                        return items;
                    }).subscribe().with(x -> {}/*no op*/);
                }
            }
        }
    }
}
