package com.acme.data.hibernatereactive.repository.support;

import com.acme.data.hibernatereactive.repository.HibernateReactiveRepository;
import io.smallrye.mutiny.Uni;
import org.hibernate.reactive.mutiny.Mutiny;
import org.reactivestreams.Publisher;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Sort;
import org.springframework.data.repository.core.RepositoryMetadata;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.Arrays;
import java.util.function.Function;
import java.util.stream.Collectors;

import static io.smallrye.mutiny.converters.uni.UniReactorConverters.toMono;

@Transactional
public class SimpleHibernateReactiveRepository<T, ID> implements HibernateReactiveRepository<T, ID> {

    private final Class<T> entityClass;
    private final String entityName;
    private final String idName;
    private final Mutiny.SessionFactory sessionFactory;
    private final Logger logger = LoggerFactory.getLogger(getClass());

    public SimpleHibernateReactiveRepository(RepositoryMetadata metadata, Mutiny.SessionFactory sessionFactory) {
        Assert.notNull(metadata, "RepositoryMetadata must not be null!");
        Assert.notNull(sessionFactory, " Mutiny.SessionFactory must not be null!");

        var entityType = sessionFactory.getMetamodel().getEntities().stream()
                .filter(et -> et.getJavaType() == metadata.getDomainType())
                .findFirst()
                .orElseThrow();

        // important metadata for this class
        this.entityClass = (Class<T>) entityType.getJavaType();
        this.entityName = entityType.getName();
        this.idName = entityType.getId(entityType.getIdType().getJavaType()).getName();
        this.sessionFactory = sessionFactory;
        this.logger.info("create repository for {}, entityName : {}, id name : {}", entityClass.getName(), entityName, idName);
    }

    @Override
    public <S extends T> Mono<S> save(S entity) {
        Assert.notNull(entity, "entity must not be null!");

        return this.sessionFactory.withSession(session -> {
                    return session.persist(entity)
                            .call(() -> session.flush())
                            .map(nothing -> entity);
                })
                .convert().with(toMono());

    }

    @Override
    public <S extends T> Flux<S> saveAll(Iterable<S> entities) {
        Assert.notNull(entities, "entities must not be null!");
        return Flux.fromIterable(entities).flatMap(e -> save(e));
    }

    @Override
    public <S extends T> Flux<S> saveAll(Publisher<S> entityStream) {
        Assert.notNull(entityStream, "entities must not be null!");
        return Flux.from(entityStream).flatMap(x->save(x));
    }

    @Override
    public Mono<T> findById(ID id) {
        Assert.notNull(id, "id must not be null!");
        return sessionFactory.withSession(session -> session.find(this.entityClass, id))
                .convert()
                .with(toMono());
    }

    @Override
    public Mono<T> findById(Publisher<ID> id) {

        Assert.notNull(id, "id must not be null!");
        return Mono.from(id).flatMap(value -> findById(value));
    }


    @Override
    public Mono<T> findById(ID id, String... fetchs) {
        Assert.notNull(id, "id must not be null!");

        String fetchClause = Arrays.stream(fetchs)
                .map(f -> "left join fetch e." + f)
                .collect(Collectors.joining(" "));

        String qlString = String.format("select e from %s e %s where e.%s=:id", this.entityName, fetchClause, this.idName);

        return sessionFactory.withSession(session -> {
                    return session.createQuery(qlString, entityClass)
                            .setParameter("id", id)
                            .getResultList();
                }).convert()
                .with(toMono())
                .flatMap(list -> list.isEmpty() ? Mono.empty() : Mono.just(list.get(0)));
    }

    @Override
    public Mono<Boolean> existsById(ID id) {
        Assert.notNull(id, "id must not be null!");
        String qlString = String.format("select e.%s from %s e where e.%s = :id", idName, entityName, idName);
        return sessionFactory.withSession(session -> {
                    return session.createQuery(qlString)
                            .setParameter("id", id)
                            .setFirstResult(0).setMaxResults(1)
                            .getResultList();
                }).convert()
                .with(toMono())
                .map(list -> !list.isEmpty());
    }

    @Override
    public Mono<Boolean> existsById(Publisher<ID> id) {
        return Mono.from(id).flatMap(value -> existsById(value));
    }

    @Override
    public Flux<T> findAll() {
        String qlString = String.format("select e from %s e", this.entityName);
        return sessionFactory.withSession(session -> {
                    return session.createQuery(qlString, this.entityClass).getResultList();
                }).convert()
                .with(toMono())
                .flatMapMany(Flux::fromIterable);
    }

    @Override
    public Flux<T> findAllById(Iterable<ID> ids) {
        Assert.notNull(ids, "ids must not be null!");
        String qlString = String.format("select e from %s e where e.%s in :ids", entityName, idName);
        return sessionFactory.withSession(session -> {
                    Mutiny.Query<T> query = session.createQuery(qlString, this.entityClass);
                    query.setParameter("ids", ids);
                    return query.getResultList();
                }).convert()
                .with(toMono())
                .flatMapMany(Flux::fromIterable);
    }

    @Override
    public Flux<T> findAllById(Publisher<ID> idStream) {
        Assert.notNull(idStream, "idStream must not be null!");
        return Flux.from(idStream).buffer()
                .filter(x -> x.isEmpty())
                .concatMap(x -> findAllById(x));
    }

    @Override
    public Mono<Long> count() {
        String qlString = String.format("select count(e) from %s e", entityName);
        return sessionFactory.withSession(session -> {
                    Mutiny.Query<Long> query = session.createQuery(qlString, Long.class);
                    return query.getSingleResult();
                }).convert()
                .with(toMono());
    }

    @Override
    public Mono<Void> deleteById(ID id) {
        Assert.notNull(id, "id must not be null!");

        return sessionFactory.withSession(session -> {
            return session.find(this.entityClass, id)
                    .call(entity -> session.remove(entity))
                    .call(() -> session.flush());
        }).convert().with(toMono()).then();
    }

    @Override
    public Mono<Void> deleteById(Publisher<ID> id) {
        Assert.notNull(id, "id must not be null!");
        return Mono.from(id).flatMap(value -> deleteById(value));
    }

    @Override
    public Mono<Void> delete(T entity) {
        Assert.notNull(entity, "entity must not be null!");

        return sessionFactory.withSession(session -> session.remove(entity)).convert().with(toMono());
    }

    @Override
    public Mono<Void> deleteAllById(Iterable<? extends ID> ids) {
        Assert.notNull(ids, "ids must not be null!");

        String qlString = String.format("delete from %s e where e.%s in :ids", this.entityName, this.idName);
        return sessionFactory.withSession(session -> {
            Mutiny.Query query = session.createQuery(qlString);
            query.setParameter("ids", ids);
            return (Uni<Integer>) query.executeUpdate();
        }).convert().with(toMono()).then();
    }

    @Override
    public Mono<Void> deleteAll(Iterable<? extends T> entities) {
        Assert.notNull(entities, "entities must not be null!");

        return Flux.fromIterable(entities).flatMap(e -> delete(e)).then();
    }

    @Override
    public Mono<Void> deleteAll(Publisher<? extends T> entityStream) {
        Assert.notNull(entityStream, "entityStream must not be null!");

        return Flux.from(entityStream).flatMap(e -> delete(e)).then();
    }

    @Override
    public Mono<Void> deleteAll() {
        String qlString = String.format("delete from %s", entityName);
        return sessionFactory.withSession(session -> {
                    return session.createQuery(qlString).executeUpdate();
                }).convert()
                .with(toMono()).then();
    }

    @Override
    public Flux<T> findAll(Sort sort) {
        String orderBy = sort.stream()
                .map(s -> String.format("e.%s %s", s.getProperty(), s.getDirection()))
                .collect(Collectors.joining(", "));

        String qlString = String.format("select e from %s e order by %s", this.entityName, orderBy);

        return sessionFactory.withSession(session -> {
                    return session.createQuery(qlString, this.entityClass).getResultList();
                }).convert()
                .with(toMono())
                .flatMapMany(Flux::fromIterable);
    }
}