package com.acme.data.hibernatereactive.sample;

import com.acme.data.hibernatereactive.repository.config.EnableHibernateReactiveRepositories;
import com.acme.data.hibernatereactive.sample.domain.Book;
import com.acme.data.hibernatereactive.sample.repository.BookRepository;
import io.vertx.pgclient.PgPool;
import io.vertx.sqlclient.SqlClient;
import io.vertx.sqlclient.Tuple;
import org.hibernate.collection.internal.PersistentBag;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.reactivestreams.Publisher;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.domain.Sort;
import org.springframework.transaction.TransactionManager;
import org.springframework.transaction.annotation.EnableTransactionManagement;
import org.springframework.util.StreamUtils;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import reactor.test.StepVerifier;
import reactor.test.StepVerifierOptions;

import java.io.IOException;
import java.nio.charset.Charset;
import java.util.List;

import static com.acme.data.hibernatereactive.sample.BookRepositoryTest.ConfigClass;
import static org.junit.jupiter.api.Assertions.*;

@SpringBootTest(classes = ConfigClass.class)
class BookRepositoryTest {

    @SpringBootApplication
    @EnableHibernateReactiveRepositories
    static class ConfigClass {
    }

    @Autowired
    private BookRepository repository;

    @Value("${spring.hibernate-reactive.properties.javax.persistence.jdbc.url}")
    private String connectionUrl;

    @BeforeEach
    void resetDb() throws IOException {
        SqlClient client = PgPool.client(this.connectionUrl);
        String script = StreamUtils.copyToString(getClass().getResourceAsStream("/reset.sql"), Charset.defaultCharset());
        client.query(script).execute().onComplete(x -> client.close())
                .toCompletionStage()
                .toCompletableFuture()
                .join();
    }

    private long getNumberOfItemsInTable() {
        SqlClient client = PgPool.client(this.connectionUrl);
        try {
            String query = "select count(*) from Book";
            return client.query(query).execute().toCompletionStage().toCompletableFuture().get()
                    .iterator()
                    .next()
                    .getLong(0);
        } catch (Exception e) {
            throw new RuntimeException(e);
        } finally {
            client.close();
        }
    }

    private boolean itemExists(Object id) {
        SqlClient client = PgPool.client(this.connectionUrl);
        try {
            String query = "select count(*) from Book where id=$1";
            return client.preparedQuery(query).execute(Tuple.of(id)).toCompletionStage().toCompletableFuture().get()
                    .iterator()
                    .next()
                    .getLong(0) > 0;
        } catch (Exception e) {
            throw new RuntimeException(e);
        } finally {
            client.close();
        }
    }

    @Test
    void findById() {
        Mono<Book> mono = repository.findById(1);
        StepVerifier.create(mono)
                .expectNextMatches(b -> b.getId() == 1)
                .verifyComplete();
    }

    @Test
    void findByIdGivenPublisher() {
        Mono<Book> mono = repository.findById(Mono.just(1));
        StepVerifier.create(mono)
                .expectNextMatches(b -> b.getId() == 1)
                .verifyComplete();
    }

    @Test
    void findByIdGivenFetch() {

        Mono<Book> mono = repository.findById(1, "reservations");
        StepVerifier.create(mono)
                .expectNextMatches(b -> ((PersistentBag)b.getReservations()).wasInitialized())
                .verifyComplete();
    }

    @Test
    void existsById() {
        Mono<Boolean> mono = repository.existsById(1);
        StepVerifier.create(mono)
                .expectNextMatches(b -> b == true)
                .verifyComplete();
    }

    @Test
    void existsByIdGivenUnknownId() {
        Mono<Boolean> mono = repository.existsById(99);
        StepVerifier.create(mono)
                .expectNextMatches(b -> b == false)
                .verifyComplete();
    }

    @Test
    void existsByIdGivenPublisher() {
        Mono<Boolean> mono = repository.existsById(Mono.just(1));
        StepVerifier.create(mono)
                .expectNextMatches(b -> b == true)
                .verifyComplete();
    }

    @Test
    void findAll() {
        long n = getNumberOfItemsInTable();
        Flux<Book> all = repository.findAll();
        StepVerifier.create(all)
                .expectNextCount(n)
                .verifyComplete();
    }

    @Test
    void findAllGivenSort() {
        Flux<Book> all = repository.findAll(Sort.by("title").descending().and(Sort.by("author.lastname")));
        StepVerifier.create(all)
                .thenConsumeWhile(x -> {
                    System.out.println(x.getTitle());
                    return true;
                })
                .verifyComplete();
    }

    @Test
    void findAllById() {
        Flux<Book> someBooks = repository.findAllById(List.of(1, 2));
        StepVerifier.create(someBooks)
                .thenConsumeWhile(x -> x.getId() == 1 || x.getId() == 2)
                .expectComplete()
                .verify();
    }

    @Test
    void findAllByIdGivenPublisher() {
        Flux<Book> someBooks = repository.findAllById(Flux.just(1, 2));
        StepVerifier.create(someBooks)
                .thenConsumeWhile(x -> x.getId() == 1 || x.getId() == 2)
                .verifyComplete();
    }

    @Test
    void count() {
        long n = getNumberOfItemsInTable();
        Mono<Long> count = repository.count();
        StepVerifier.create(count)
                .expectNextMatches(x -> x == n)
                .verifyComplete();
    }

    @Test
    void deleteById() {
        long n = getNumberOfItemsInTable();
        Mono<Void> mono = repository.deleteById(3);
        StepVerifier.create(mono).verifyComplete();
        assertEquals(n - 1, getNumberOfItemsInTable());
        assertFalse(itemExists(3));
    }

    @Test
    void deleteByIdGivenPublisherOfIds() {
        long n = getNumberOfItemsInTable();
        Mono<Void> mono = repository.deleteById(Mono.just(3));
        StepVerifier.create(mono).verifyComplete();
        assertEquals(n - 1, getNumberOfItemsInTable());
        assertFalse(itemExists(3));
    }

    @Test
    void delete() {
        long n = getNumberOfItemsInTable();
        Book b = this.repository.findById(3).block();
        Mono<Void> mono = this.repository.delete(b);
        StepVerifier.create(mono).verifyComplete();
        assertEquals(n - 1, getNumberOfItemsInTable());
        assertFalse(itemExists(3));
    }

    @Test
    void deleteAllByIds() {
        long n = getNumberOfItemsInTable();
        Mono<Void> mono = repository.deleteAllById(List.of(3, 4));
        StepVerifier.create(mono).verifyComplete();
        assertEquals(n - 2, getNumberOfItemsInTable());
        assertFalse(itemExists(3));
        assertFalse(itemExists(4));
    }

    @Test
    void deleteAllGivenEntities() {
        long n = getNumberOfItemsInTable();
        Iterable<Book> books = this.repository.findAllById(List.of(3, 4)).toIterable();
        Mono<Void> mono = this.repository.deleteAll(books);
        StepVerifier.create(mono).verifyComplete();
        assertEquals(n - 2, getNumberOfItemsInTable());
        assertFalse(itemExists(3));
        assertFalse(itemExists(4));
    }

    @Test
    void deleteAllEntitiesGivenEntityStream() {
        long n = getNumberOfItemsInTable();
        Publisher<Book> entityStream = this.repository.findAllById(List.of(3, 4));
        Mono<Void> mono = this.repository.deleteAll(entityStream);
        StepVerifier.create(mono).verifyComplete();
        assertEquals(n - 2, getNumberOfItemsInTable());
        assertFalse(itemExists(3));
        assertFalse(itemExists(4));
    }

    @Test
    void deleteAll() {
        Mono<Void> mono = repository.deleteAll();
        StepVerifier.create(mono).verifyComplete();
        assertEquals(0, getNumberOfItemsInTable());
    }

    @Test
    void save() {
        long n = getNumberOfItemsInTable();
        var book = new Book();
        book.setTitle("new book");
        Mono<Book> mono = this.repository.save(book);
        StepVerifier.create(mono)
                .expectNextMatches(x -> x.getId() != null)
                .verifyComplete();
        assertEquals(n + 1, getNumberOfItemsInTable());
        assertTrue(itemExists(book.getId()));
    }

    @Test
    void saveAllGivenEntities() {
        var book1 = new Book();
        book1.setTitle("new book");
        var book2 = new Book();
        book2.setTitle("another new book");

        List<Book> newBooks = List.of(book1, book2);
        long n = getNumberOfItemsInTable();
        Flux<Book> flux = this.repository.saveAll(newBooks).cache();
        StepVerifier.create(flux)
                .thenConsumeWhile(x -> x.getId() != null)
                .verifyComplete();
        assertEquals(n + newBooks.size(), getNumberOfItemsInTable());
        flux.toIterable().forEach(b -> assertTrue(itemExists(b.getId())));
    }

    @Test
    void saveAllGivenEntityStream() {
        var book1 = new Book();
        book1.setTitle("new book");
        var book2 = new Book();
        book2.setTitle("another new book");

        List<Book> newBooks = List.of(book1, book2);

        long n = getNumberOfItemsInTable();
        Flux<Book> flux = this.repository.saveAll(Flux.fromIterable(newBooks)).cache();
        StepVerifier.create(flux)
                .thenConsumeWhile(x -> x.getId() != null)
                .verifyComplete();
        assertEquals(n + newBooks.size(), getNumberOfItemsInTable());
        flux.toIterable().forEach(b -> assertTrue(itemExists(b.getId())));
    }
}